<?php

namespace Drupal\sdc_library_paragraphs\Form;

use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Extension\ThemeExtensionList;
use Drupal\Core\Field\WidgetBase;
use Drupal\Core\Form\FormInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Theme\ComponentPluginManager;
use Drupal\Core\Theme\ThemeInitialization;
use Drupal\Core\Theme\ThemeManagerInterface;
use Drupal\sdc_library\ComponentManagerInterface;
use Drupal\sdc_library\Form\ComponentForm;
use Drupal\sdc_library_paragraphs\ParagraphsManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Yaml\Yaml;

/**
 * Extend SDC Library component form to support Paragraphs.
 */
class ParagraphsForm extends ComponentForm implements FormInterface {

  /**
   * {@inheritdoc}
   */
  public function __construct(
    protected EntityTypeManager $entityTypeManager,
    protected ThemeManagerInterface $themeManager,
    protected ThemeInitialization $themeInitialization,
    protected ThemeExtensionList $themeExtensionList,
    protected $request_stack,
    protected $renderer,
    protected $config_factory,
    public string $sdcWelcomeComponent,
    protected ComponentManagerInterface $componentManager,
    protected ComponentPluginManager $componentPluginManager,
    protected ModuleHandlerInterface $module_handler,
    protected ParagraphsManagerInterface $sdcLibraryParagraphsManager,
  ) {
    $this->requestStack = $request_stack;
    $this->configFactory = $config_factory;
    $this->moduleHandler = $module_handler;
    $this->sdcLibraryParagraphsManager = $sdcLibraryParagraphsManager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('theme.manager'),
      $container->get('theme.initialization'),
      $container->get('extension.list.theme'),
      $container->get('request_stack'),
      $container->get('renderer'),
      $container->get('config.factory'),
      $welcome_component = (string) $container->getParameter('sdc_library.welcome_component'),
      $container->get('sdc_library.component_manager'),
      $container->get('plugin.manager.sdc'),
      $container->get('module_handler'),
      $container->get('sdc_library_paragraphs.paragraphs_manager'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'sdc_library_paragraphs';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    // We are extending the parent form.
    $form = parent::buildForm($form, $form_state);
    // This is where we check if the component has been matched to a paragraph.
    $component_yml = Yaml::parseFile(DRUPAL_ROOT . '/' . $this->activeComponent->metadata->path . '/' . $this->activeComponent->metadata->machineName . '.component.yml');
    $use_paragraph = $component_yml['use_paragraph'] ?? FALSE;
    // The $use_paragraph variable is the machine name of the paragraph to use.
    if ($use_paragraph) {
      // We do not need the SDC's props and slots anymore so get rid of them.
      unset($form['props']);
      unset($form['slots']);

      // If this #parents doesn't get set we see a WSOD.
      $form['#parents'] = [];

      // Check if the form is being submitted and that we have parent ID.
      if ($form_state->getTriggeringElement() && $form_state->getValue('_parent_entity_id')) {
        $sdc_parent_entity_id = $form_state->getValue('_parent_entity_id');
        /** @var \Drupal\sdc_library_paragraphs\Entity\SdcLibraryParagraphs $sdc_parent_entity */
        $sdc_parent_entity = $this->entityTypeManager->getStorage('sdc_library_paragraphs')->load($sdc_parent_entity_id);
        // Look for chosen theme or return an empty string to avoid TypeError.
        $this->chosenTheme = $form_state->getValue('_sdcChosentheme') ?? '';
        // Thank you Paragraphs Previewer module for this WidgetBase code.
        $field_parents = [];
        $widget_state = WidgetBase::getWidgetState($field_parents, 'field_sdc_referenced_paragraph', $form_state);
        // We can hard-code 0 (zero) because our SDC Library Paragraph entity's
        // field_sdc_referenced_paragraph field has a cardinality of 1.
        if (!empty($widget_state['paragraphs'][0]['entity'])) {
          $paragraph = $widget_state['paragraphs'][0]['entity'];
          // Create a new paragraph with no id.
          $paragraph_clone = $paragraph->createDuplicate();
          // Set and save the entity. This is what we are going to render.
          $sdc_parent_entity->set('field_sdc_referenced_paragraph', $paragraph_clone);
        }
      }
      else {
        // Build the containing parent entity. It's where the paragraph lives.
        /** @var \Drupal\sdc_library_paragraphs\Entity\SdcLibraryParagraphs $sdc_parent_entity */
        $sdc_parent_entity = $this->entityTypeManager->getStorage('sdc_library_paragraphs')->create(
          [
            'title' => $this->t('SDC Library Paragraph Entity: :paragraph (timestamp :timestamp)',
              [
                ':paragraph' => $use_paragraph,
                ':timestamp' => time(),
              ]
            ),
          ]
        );
        // Check if prefill config is toggled. If so, we look for an existing
        // paragraph and prefill the form with its data, to save user effort.
        $use_paragraph_prefill = $this->configFactory->get('sdc_library_paragraphs.settings')->get('prefill');
        if ($use_paragraph_prefill) {
          // Track which arbitrary paragraph is being used to prefill the form.
          $paragraph_target_id = $this->sdcLibraryParagraphsManager->getArbitraryParagraphId($use_paragraph);
          // It's not guaranteed the site has an existing paragraph to use.
          if ($paragraph_target_id) {
            $sdc_reference_paragraph = $this->entityTypeManager->getStorage('paragraph')->load($paragraph_target_id);
            $paragraph_clone = $sdc_reference_paragraph->createDuplicate();
            // Set and save the entity. This is what we are going to render.
            $sdc_parent_entity->set('field_sdc_referenced_paragraph', $paragraph_clone);
          }
          else {
            // This paragraph type has no instances yet so create an empty form.
            $sdc_reference_paragraph = $this->entityTypeManager->getStorage('paragraph')->create(
              [
                'type' => $use_paragraph,
                'parent_type' => $sdc_parent_entity->getEntityTypeId(),
              ]
            );
            $sdc_parent_entity->set('field_sdc_referenced_paragraph', $sdc_reference_paragraph);
          }
        }
      }

      // At this point we definitely have a parent entity to save.
      $sdc_parent_entity->save();
      // Add the parent entity id as a hidden form input.
      $form['_parent_entity_id'] = [
        '#type' => 'hidden',
        '#value' => $sdc_parent_entity->id(),
      ];

      // The $form_display and $field_widget variables build out paragraph form.
      /** @var \Drupal\Core\Entity\Display\EntityFormDisplayInterface $form_display */
      $form_display = $this->entityTypeManager->getStorage('entity_form_display')->load('sdc_library_paragraphs.sdc_library_paragraphs.default');
      /** @var \Drupal\paragraphs\Plugin\Field\FieldWidget\ParagraphsWidget $field_widget */
      $field_widget = $form_display->getRenderer('field_sdc_referenced_paragraph');
      $items = $sdc_parent_entity->get('field_sdc_referenced_paragraph');
      // This is the paragraph form built for us by Drupal.
      $form['field_sdc_referenced_paragraph'] = $field_widget->form($items, $form, $form_state);

      // We post this form to the page contained in the iframe.
      $form['#action'] = '/sdc_library/component/' . $this->sdcPluginId . '/' . $this->chosenTheme . '/' . $sdc_parent_entity->id();
      // Use a theme template to build the iframe in the form. The iframe's src
      // is the same as the form action.
      $form['sdc_library_iframe'] = [
        '#theme' => 'sdc_library__iframe',
        '#src' => $form['#action'],
        '#weight' => -50,
      ];

      // Use ajax callback to submit and render the paragraph.
      $form['submit'] = [
        '#type' => 'submit',
        '#value' => self::SUBMIT_BTN_TEXT,
        '#weight' => -30,
        '#submit' => ['::ajaxUpdateIframe'],
        '#ajax' => [
          'callback' => '::ajaxUpdateIframe',
          'wrapper' => 'sdc-iframe-wrapper',
          'progress' => [
            'type' => 'throbber',
            'message' => $this->t('Rendering paragraph...'),
          ],
        ],
      ];
    }
    return $form;
  }

  /**
   * Ajax method to update iframe source.
   */
  public function ajaxUpdateIframe(array &$form, FormStateInterface $form_state) {
    // We need setRebuild to access $form_state in WidgetBase::getWidgetState().
    $form_state->setRebuild();
    // Build the new src to pass to the iframe.
    $iframe_src = '/sdc_library/component/' . $this->sdcPluginId . '/' . $this->chosenTheme . '/' . $form_state->getValue('_parent_entity_id');
    // Build and return the iframe's render array.
    $build = [
      '#theme' => 'sdc_library__iframe',
      '#prefix' => '<div id="sdc-iframe-wrapper">',
      '#suffix' => '</div>',
      '#src' => $iframe_src,
    ];
    return $build;
  }

}
