<?php

namespace Drupal\sdc_library_paragraphs;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Asset\AssetQueryString;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\State\StateInterface;
use Drupal\Core\Theme\ComponentPluginManager;
use Drupal\Core\Theme\ThemeManagerInterface;
use Drupal\sdc_library\ComponentManager;
use Drupal\sdc_library_paragraphs\Form\ParagraphsForm;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\Yaml\Yaml;

/**
 * SDC Library Paragraphs manager service.
 */
class ParagraphsManager extends ComponentManager implements ParagraphsManagerInterface {

  /**
   * Constructor.
   */
  public function __construct(
    protected ComponentManager $componentManager,
    protected RequestStack $requestStack,
    protected $formBuilder,
    protected ComponentPluginManager $componentPluginManager,
    protected ThemeManagerInterface $themeManager,
    public string $sdcWelcomeComponent,
    public string $sdcIframeWrapperId,
    public string $sdcRenderedComponentWrapperId,
    protected AssetQueryString $assetQueryString,
    protected TimeInterface $time,
    protected StateInterface $state,
    protected $renderer,
    protected EntityTypeManagerInterface $entityTypeManager,
    protected ConfigFactoryInterface $configFactory,
  ) {
  }

  /**
   * {@inheritdoc}
   */
  public function renderComponentForm(): array {
    // The settings file has a toggle to disable paragraphs functionality.
    $is_enabled = (bool) $this->configFactory->get('sdc_library_paragraphs.settings')->get('enabled');
    if ($is_enabled) {
      return $this->formBuilder->getForm(ParagraphsForm::class);
    }
    // This returns control back to the regular component form.
    return parent::renderComponentForm();
  }

  /**
   * {@inheritdoc}
   */
  public function renderComponent(): array {
    // This changes the query string so that fresh css/js is loaded each time.
    $this->assetQueryString->reset();
    $request = $this->requestStack->getCurrentRequest();
    if ($request->get('sdcid')) {
      return $this->getSdcEntityRenderArray($request->get('sdcid'));
    }
    // We give the user the option to prefill data from an existing paragraph.
    $use_paragraph_prefill = $this->configFactory->get('sdc_library_paragraphs.settings')->get('prefill');
    $paragraph_machine_name = $this->getParagraphMachineNameFromActiveComponent();
    if ($use_paragraph_prefill && $paragraph_machine_name) {
      $paragraph_target_id = $this->getArbitraryParagraphId($paragraph_machine_name);
      if ($paragraph_target_id) {
        return $this->getSpecificParagraphRenderArray($paragraph_target_id);
      }
    }
    // Looks like we do not have a paragraph to render. Call the parent method.
    return $this->componentManager->renderComponent();
  }

  /**
   * {@inheritdoc}
   */
  public function getSdcEntityRenderArray(mixed $sdcid): array {
    // Instantiate a render array that gets overwritten if we have an $sdcid.
    $rendered_component = [
      '#type' => 'markup',
      '#markup' => $this->t('<h3 class="sdc_library_paragraphs__empty">:blurb</h3>', [':blurb' => 'There is no paragraph to render.']),
    ];
    if ($sdcid) {
      /** @var \Drupal\sdc_library_paragraphs\Entity\SdcLibraryParagraphs $sdc_parent_entity */
      $sdc_parent_entity = $this->entityTypeManager->getStorage('sdc_library_paragraphs')->load($sdcid);
      $elements = $sdc_parent_entity->field_sdc_referenced_paragraph->view('default');
      $rendered_component = [
        '#type' => 'markup',
        '#markup' => $this->renderer->render($elements),
      ];
    }
    return $rendered_component;
  }

  /**
   * {@inheritdoc}
   */
  public function getParagraphMachineNameFromActiveComponent(): ?string {
    $activeComponent = $this->componentManager->getActiveComponent();
    $component_yml = Yaml::parseFile(DRUPAL_ROOT . '/' . $activeComponent->metadata->path . '/' . $activeComponent->metadata->machineName . '.component.yml');
    return $component_yml['use_paragraph'] ?? NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function getArbitraryParagraphId(string $paragraph_machine_name = ''): ?int {
    if ($paragraph_machine_name) {
      $possible_paragraphs = $this->entityTypeManager->getStorage('paragraph')->getQuery()
        ->condition('type', $paragraph_machine_name, '=')
        ->condition('parent_type', 'sdc_library_paragraphs', '<>')
        ->sort('id', 'DESC')
        ->accessCheck(TRUE)
        ->execute();
      if (!empty($possible_paragraphs)) {
        // Use the first item which is the most recent paragraph of this type.
        return (int) array_shift($possible_paragraphs);
      }
    }
    return NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function getSpecificParagraphRenderArray(int $paragraph_target_id): array {
    // Instantiate a render array in case we do not find a paragraph to use.
    $rendered_component = [
      '#type' => 'markup',
      '#markup' => $this->t('<h3 class="sdc_library_paragraphs__empty">:blurb</h3>', [':blurb' => 'We could not find a paragraph with which to prefill data.']),
    ];
    if ($paragraph_target_id) {
      $paragraph = $this->entityTypeManager->getStorage('paragraph')->load($paragraph_target_id);
      $view_builder = $this->entityTypeManager->getViewBuilder($paragraph->getEntityTypeId());
      $rendered_component = $view_builder->view($paragraph);
    }
    return $rendered_component;
  }

}
