<?php

namespace Drupal\sdc_library;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Asset\AssetQueryString;
use Drupal\Core\Plugin\Component;
use Drupal\Core\State\StateInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Theme\ComponentPluginManager;
use Drupal\Core\Theme\ThemeManagerInterface;
use Drupal\Core\Url;
use Drupal\sdc\Exception\ComponentNotFoundException;
use Drupal\sdc_library\Form\ComponentForm;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\Yaml\Yaml;

/**
 * Component manager service.
 */
class ComponentManager implements ComponentManagerInterface {

  // This gives access to $this->t().
  use StringTranslationTrait;

  /**
   * Constructor.
   */
  public function __construct(
    protected RequestStack $requestStack,
    protected $formBuilder,
    protected ComponentPluginManager $componentPluginManager,
    protected ThemeManagerInterface $themeManager,
    public string $sdcWelcomeComponent,
    public string $sdcIframeWrapperId,
    public string $sdcRenderedComponentWrapperId,
    protected AssetQueryString $assetQueryString,
    protected TimeInterface $time,
    protected StateInterface $state,
    protected $renderer,
  ) {
  }

  /**
   * {@inheritdoc}
   */
  public function getComponentsTree(): ?array {
    // We build a nested tree structure of the list of components.
    $request = $this->requestStack->getCurrentRequest();
    $components = $this->componentPluginManager->getAllComponents();
    // Instantiate variable to hold site components, grouped by theme/module.
    $items = [];
    foreach ($components as $component) {
      $component_yml = Yaml::parseFile(DRUPAL_ROOT . '/' . $component->metadata->path . '/' . $component->metadata->machineName . '.component.yml');
      $sdc_library_hide = $component_yml['sdc_library_hide'] ?? FALSE;
      if (!$sdc_library_hide) {
        // This grouping approach is taken from the sdc_styleguide module.
        $definition = $component->getPluginDefinition();
        $group_key_name = strtolower("{$definition['extension_type']->name}:{$definition['provider']}");
        // @todo Not the slickest, all we do is add a space between the "name"
        // and the "provider" values so it's easier for the user to read.
        $group_display_name = "{$definition['extension_type']->name}: {$definition['provider']}";
        if (!isset($items[$group_key_name])) {
          $items[$group_key_name] = [];
          $items[$group_key_name]['id'] = $definition['id'];
          $items[$group_key_name]['title'] = $group_display_name;
          $items[$group_key_name]['name'] = $group_display_name;
        }
        // We need to provide route parameters to build our $url variable.
        $route_parameters = [
          'component_plugin' => $component->getPluginId(),
          'chosen_theme' => $request->get('chosen_theme') ?? $this->themeManager->getActiveTheme()->getName(),
        ];

        $url = Url::fromRoute('sdc_library.browse', $route_parameters);
        // It is in the "below" array that we add this group's components.
        // The "below" key is used in twig to build out menu-like structure.
        $items[$group_key_name]['below'][] = [
          'url' => $url,
          'href' => $url->toString(),
          'title' => $component->metadata->name,
          'name' => $component->metadata->name,
          'status' => $component->metadata->status,
          'group' => $component->metadata->group,
          'extension' => $this->findExtensionName($component->metadata->path),
        ];
        // This sorts the components within their parent groupings. My gut says
        // it is wrong to sort on each loop iteration, but it works!
        usort($items[$group_key_name]['below'], $this->sortByComponentName(...));
      }
    }
    // This sorts the parent groupings.
    usort($items, $this->sortByComponentName(...));

    // Let's shift the SDC Library components group to the beginning of items.
    $sdc_library_index = array_search('sdc_library:welcome', array_column($items, 'id'));
    // Grab the contents of our SDC Library group so we can unset it and add it
    // back with array_unshift(). There's likely a more eloquent way to do this.
    $temp_array_holder = $items[$sdc_library_index];
    unset($items[$sdc_library_index]);
    array_unshift($items, $temp_array_holder);

    return $items;
  }

  /**
   * {@inheritdoc}
   */
  public function renderComponentList(): array {
    // This gets called directly from the BrowseLibraryController controller.
    $items = $this->getComponentsTree();
    // Early return with warning message if we find no components.
    if (!$items) {
      return [
        '#markup' => '<div class="messages messages--warning"><h3>' . $this->t('Unable to find any components') . '</h3>' . $this->t('Check that the modules or themes containing the components are enabled and the components are correctly named.</div>'),
      ];
    }
    $build = [
      '#theme' => 'sdc_list',
      '#items' => $items,
      '#component_form' => $this->renderComponentForm(),
    ];
    return $build;

  }

  /**
   * {@inheritdoc}
   */
  public function renderComponentForm(): array {
    return $this->formBuilder->getForm(ComponentForm::class);
  }

  /**
   * {@inheritdoc}
   */
  public function renderComponent(): array {
    try {
      $build = $this->generateRenderArray(
        $this->getActiveComponent()
      );
    }
    catch (ComponentNotFoundException $e) {
      $build = [
        '#markup' => '<div class="messages messages--error"><h3>' . $this->t('Unable to find that component') . '</h3>' . $this->t('Check that the module or theme containing the component is enabled and the component is correctly named. Message: %message', ['%message' => $e->getMessage()]) . '</div>',
      ];
    }
    // This changes the query string so that fresh css/js is loaded each time.
    $this->assetQueryString->reset();
    return $build;
  }

  /**
   * {@inheritdoc}
   */
  public function getActiveComponent(): Component {
    // Clearing out the cache to keep things fresh. During dev, this allowed me
    // to update default values in blockquote.component.yml and the form ajax
    // would update them immediately.
    $this->componentPluginManager->clearCachedDefinitions();

    $request = $this->requestStack->getCurrentRequest();
    if (!empty($request->get('component_plugin'))) {
      return $this->componentPluginManager->createInstance($request->get('component_plugin'));
    }
    // In case the above does not run, return our default Welcome component.
    // @todo This string concatenation is janky.
    return $this->componentPluginManager->createInstance('sdc_library:' . $this->sdcWelcomeComponent);
  }

  /**
   * {@inheritdoc}
   */
  public function getPropsFromComponentMetadata(array $schema_properties): ?array {
    $props = [];
    foreach ($schema_properties as $property => $property_data) {
      // The property type is always in an array but we want a scalar value.
      $type = array_shift($property_data['type']);
      // We hope that each property has been given a default, but if not, we try
      // to figure out something sensible to use. In the case of properties with
      // "enum" and no default, we can just grab the first enum value.
      if (isset($type)) {
        // This crucial bit is where we determine what to pass as the default.
        if (isset($property_data['default'])) {
          $props[$property] = $this->getPropertyValue($type, $property_data['default']);
        }
        elseif (isset($property_data['examples'])) {
          // Using array_shift to get the first value as there could be many.
          $props[$property] = array_shift($property_data['examples']);
        }
        elseif (isset($property_data['enum']) && !isset($property_data['default'])) {
          // Just grab the first value from the enum list to use as default.
          $props[$property] = array_shift($property_data['enum']);
        }
        else {
          // We have no defaults to use so let's point that out.
          $props[$property] = self::PROPERTY_TYPE_DEFAULTS[$type];
        }
      }
    }
    return $props;
  }

  /**
   * {@inheritdoc}
   */
  public function getSlotsFromComponentMetadata(array $metadata_slots): ?array {
    $slots = [];
    foreach ($metadata_slots as $slot => $slot_data) {
      // We assume there are examples, use the first one.
      if (isset($slot_data['examples'][0])) {
        $slots[$slot] = [
          '#type' => 'processed_text',
          '#format' => 'full_html',
          '#text' => $slot_data['examples'][0],
        ];
      }
    }
    return $slots;
  }

  /**
   * {@inheritdoc}
   */
  public function getPropsFromPostedForm(Component $component, array $posted_data): ?array {
    $props = [];
    if (isset($posted_data['props'][$component->getPluginId()])) {
      foreach ($posted_data['props'][$component->getPluginId()] as $property => $property_data) {
        // The form returns a single value for the type.
        $type = $property_data['type'];
        if (isset($type) && isset($property_data['default'])) {
          $props[$property] = $this->getPropertyValue($type, $property_data['default']);
        }
      }
    }
    return $props;
  }

  /**
   * {@inheritdoc}
   */
  public function getSlotsFromPostedForm(Component $component, array $posted_data): ?array {
    $slots = [];
    if (isset($posted_data['slots'][$component->getPluginId()])) {
      foreach ($posted_data['slots'][$component->getPluginId()] as $slot => $slot_data) {
        // We use array_shift() because the slot may have multiple examples
        // which means they are indexed with 0, 1, 2 etc.
        $slot_data = array_shift($slot_data);
        // Return a render array, otherwise markup gets escaped.
        $slots[$slot] = [
          '#type' => 'processed_text',
          '#format' => $slot_data['format'],
          '#text' => $slot_data['value'],
        ];
      }
    }
    return $slots;
  }

  /**
   * {@inheritdoc}
   */
  public function getPropertyValue(string $type, mixed $property_value): mixed {
    // The type can be any number of types: string, number, integer, array etc.
    switch ($type) {
      case 'array':
      case 'object':
        // The component's array and object properties gives us an array for
        // $property_value whereas the submitted form gives a string.
        if (gettype($property_value) == 'string') {
          // If it is a string, assume it is json encoded so decode it.
          $property_value = json_decode($property_value);
        }
        $default_value = $property_value;
        break;

      case 'boolean':
        $default_value = boolval($property_value);
        break;

      case 'integer':
        $default_value = (int) $property_value;
        break;

      case 'null':
        // Hey, if it's a null type then we can just pass NULL, no?
        $default_value = NULL;
        break;

      case 'number':
        $default_value = (float) $property_value;
        break;

      case 'string':
        // Create a render array, for the string may contain legitimate markup.
        $render_array = [
          '#type' => 'processed_text',
          '#format' => 'full_html',
          '#text' => $property_value,
        ];
        // WYSIWYG fields definitely have markup and have this array structure.
        if (
          is_array($property_value)
          && isset($property_value['value'])
          && $property_value['format']
        ) {
          $render_array = [
            '#type' => 'processed_text',
            '#format' => $property_value['format'],
            '#text' => $property_value['value'],
          ];
        }
        $default_value = $this->renderer->render($render_array);
        break;

      default:
        $default_value = $property_value;
        break;
    }
    return $default_value;
  }

  /**
   * {@inheritdoc}
   */
  public function generateRenderArray(Component $component): array {
    $this->componentPluginManager->clearCachedDefinitions();
    /** @var \Drupal\sdc\Component\ComponentMetadata $metadata */
    $metadata = $component->metadata;
    // The properties can be built from the SDC's yml data or from POST data.
    // Check which one, get the data we need in a normalized way and then
    // continue to build the render array.
    if ($this->requestStack->getCurrentRequest()->request->all()) {
      // This is structured as the POSTed form.
      $props = $this->getPropsFromPostedForm($component, $this->requestStack->getCurrentRequest()->request->all());
      $slots = $this->getSlotsFromPostedForm($component, $this->requestStack->getCurrentRequest()->request->all());
    }
    else {
      // A regular page load gives us a PHP array structure of the SDC's yml.
      $props = $this->getPropsFromComponentMetadata($metadata->schema['properties']);
      $slots = $this->getSlotsFromComponentMetadata($metadata->slots);
    }

    $rendered_component = [
      '#type' => 'component',
      '#component' => $component->getPluginId(),
      '#props' => $props,
      '#slots' => $slots,
    ];

    $build = [
      '#theme' => 'sdc_component',
      '#rendered_component' => [
        '#type' => 'processed_text',
        '#format' => 'full_html',
        '#text' => $this->renderer->render($rendered_component),
      ],
    ];
    return $build;
  }

  /**
   * {@inheritdoc}
   */
  public function findExtensionName(string $path): ?string {
    if (empty($path)) {
      return NULL;
    }
    $path = dirname($path);
    $dir = basename($path);
    $info_file = $path . DIRECTORY_SEPARATOR . "$dir.info.yml";
    if (file_exists($info_file)) {
      return $dir;
    }
    return $this->findExtensionName($path);
  }

  /**
   * Called from usort() to order component items.
   *
   * @param array $a
   *   Array holding first name to compare.
   * @param array $b
   *   Array holding second name to compare.
   *
   * @return int
   *   Integer of -1, 0, 1 to indicate if first string is less, equal or more.
   */
  public function sortByComponentName(array $a, array $b): int {
    return strcmp($a['name'], $b['name']);
  }

}
