<?php

namespace Drupal\sdc_library;

use Drupal\Core\Plugin\Component;

/**
 * Provides an interface for the SDC Library component manager.
 */
interface ComponentManagerInterface {

  /**
   * A lookup map of property type (string, number etc) to fallback defaults.
   */
  final public const PROPERTY_TYPE_DEFAULTS = [
    'array' => [
      'foo',
      'bar',
      'baz',
    ],
    'boolean' => 1,
    'integer' => 3,
    'null' => NULL,
    'number' => 3.14,
    // @todo Not sure of the best way to handle objects.
    'object' => '{ foo: bar}',
    'string' => 'This is fallback text defined in the SDC Library module. Please add defaults to your component so that we can use them to render the component.',
  ];

  /**
   * Return a render array of all the listed components.
   */
  public function renderComponentList(): array;

  /**
   * Return the full component display of component and form.
   */
  public function renderComponentForm(): array;

  /**
   * Return component render array.
   */
  public function renderComponent(): array;

  /**
   * Return the currently active component.
   */
  public function getActiveComponent(): Component;

  /**
   * Generates a render array of the component with the its slots and props.
   *
   * @param \Drupal\Core\Plugin\Component $component
   *   The SDC component being used.
   *
   * @return array
   *   The paragraph's render array.
   */
  public function generateRenderArray(Component $component): array;

  /**
   * List site components grouped by module / theme, then sorted alphabetically.
   */
  public function getComponentsTree(): ?array;

  /**
   * Get the default component props from its component.yml.
   *
   * This gets us the defaults needed to display the component on initial page
   * load.
   *
   * @param array $schema_properties
   *   The SDC component's schema.
   *
   * @return null|array
   *   The component's props, if they exist.
   */
  public function getPropsFromComponentMetadata(array $schema_properties): ?array;

  /**
   * Get the default component slots from its component.yml.
   *
   * @param array $schema_properties
   *   The SDC component's schema.
   *
   * @return null|array
   *   The component's slots, if they exist.
   */
  public function getSlotsFromComponentMetadata(array $schema_properties): ?array;

  /**
   * Get the component props from user-entered POST data.
   *
   * This is where the user has updated default values as they see fit.
   *
   * @param \Drupal\Core\Plugin\Component $component
   *   The SDC component being used.
   * @param array $posted_data
   *   The data from the posted form.
   *
   * @return null|array
   *   The posted data to use to render the component props.
   */
  public function getPropsFromPostedForm(Component $component, array $posted_data): ?array;

  /**
   * Get the component slots from user-entered POST data.
   *
   * This is where the user has updated default values as they see fit.
   *
   * @param \Drupal\Core\Plugin\Component $component
   *   The SDC component being used.
   * @param array $posted_data
   *   The data from the posted form.
   *
   * @return null|array
   *   The posted data to use to render the component slots.
   */
  public function getSlotsFromPostedForm(Component $component, array $posted_data): ?array;

  /**
   * Return a given property's value in a normalized way for its type.
   *
   * @param string $type
   *   The property type.
   * @param mixed $property_value
   *   The property value we need to normalize.
   *
   * @return mixed
   *   The normalized value for that property type.
   */
  public function getPropertyValue(string $type, mixed $property_value): mixed;

}
