<?php

namespace Drupal\sdc_library\Form;

use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Extension\ThemeExtensionList;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\Component;
use Drupal\Core\Theme\ComponentPluginManager;
use Drupal\Core\Theme\ThemeInitialization;
use Drupal\Core\Theme\ThemeManagerInterface;
use Drupal\sdc_library\ComponentManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Build component browser form.
 */
class ComponentForm extends FormBase implements FormInterface {

  // @todo What are the best text field/area sizes/lengths to use?
  const TEXTFIELD_SIZE = 100;
  const TEXTFIELD_MAXLENGTH = 512;
  const SUBMIT_BTN_TEXT = 'Update component';

  /**
   * Hold the active component.
   *
   * @var \Drupal\Core\Plugin\Component
   */
  protected Component $activeComponent;

  /**
   * Property that holds the SDC Plugin ID. E.g. sdc_library:welcome.
   *
   * @var string
   */
  protected string $sdcPluginId = '';

  /**
   * Property that holds the site's front-end (default) theme name.
   *
   * @var string
   */
  protected string $defaultTheme = '';

  /**
   * Property that holds the chosen theme's name.
   *
   * @var string
   */
  protected string $chosenTheme = '';

  /**
   * The request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * SDC List form constructor.
   */
  public function __construct(
    protected EntityTypeManager $entityTypeManager,
    protected ThemeManagerInterface $themeManager,
    protected ThemeInitialization $themeInitialization,
    protected ThemeExtensionList $themeExtensionList,
    protected $request_stack,
    protected $renderer,
    protected $config_factory,
    public string $sdcWelcomeComponent,
    protected ComponentManagerInterface $componentManager,
    protected ComponentPluginManager $componentPluginManager,
    protected ModuleHandlerInterface $module_handler,
  ) {
    $this->requestStack = $request_stack;
    $this->configFactory = $config_factory;
    $this->moduleHandler = $module_handler;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('theme.manager'),
      $container->get('theme.initialization'),
      $container->get('extension.list.theme'),
      $container->get('request_stack'),
      $container->get('renderer'),
      $container->get('config.factory'),
      $welcome_component = (string) $container->getParameter('sdc_library.welcome_component'),
      $container->get('sdc_library.component_manager'),
      $container->get('plugin.manager.sdc'),
      $container->get('module_handler'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'sdc_library_component';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    // We populate available $themes and also use the $request object.
    $themes = $this->themeExtensionList->reset()->getAllAvailableInfo();
    $request = $this->requestStack->getCurrentRequest();

    // Instantiate any variables that get used in the form.
    $sdcPluginId = '';

    // Grab defaults and then allow route parameters to override them.
    // The default "welcome" component is defined in sdc_library.services.yml.
    $this->sdcPluginId = 'sdc_library:' . $this->sdcWelcomeComponent;
    // Instantiate default (front-end) theme name.
    $this->defaultTheme = $this->configFactory->get('system.theme')->get('default');
    // We assume the user wants to see the SDC in their front-end theme.
    $this->chosenTheme = $this->defaultTheme;
    if (!empty($request->get('component_plugin')) && !empty($request->get('chosen_theme'))) {
      $this->sdcPluginId = $request->get('component_plugin');
      // Our route's chosen_theme argument placeholder is "admin" and its
      // presence signals us to use the front-end theme. This will bump heads
      // with the edge case of an admin theme with the machine-name "admin".
      $this->chosenTheme = ($request->get('chosen_theme') == 'admin') ? $this->defaultTheme : $request->get('chosen_theme');
    }

    // Component name is easier to grok than sdcPuginId.
    $component_name = $this->sdcPluginId;
    // Load the component object to get at its properties.
    $this->activeComponent = $this->componentPluginManager->createInstance($this->sdcPluginId);

    // Nest the form values.
    $form['#tree'] = TRUE;
    // This is the iframe name parameter.
    $form['#attributes'] = [
      'target' => 'sdc-library-iframe',
    ];
    // We post this form to the page contained in the iframe.
    $form['#action'] = '/sdc_library/component/' . $this->sdcPluginId . '/' . $this->chosenTheme;
    // Use a theme template to build the iframe in the form. The iframe's src is
    // the same as the form action.
    $form['sdc_library_iframe'] = [
      '#theme' => 'sdc_library__iframe',
      '#src' => $form['#action'],
      '#weight' => -50,
    ];

    // We create a select list of the available themes.
    $options = [];
    foreach ($themes as $key => $theme) {
      $options[$key] = $theme['name'];
    }
    asort($options);

    // Default the select list to the chosen theme.
    $form['_sdcChosentheme'] = [
      '#title' => $this->t('Change theme'),
      '#type' => 'select',
      '#default_value' => $this->chosenTheme,
      '#options' => $options,
      '#weight' => -40,
    ];

    $form['_sdcPluginId'] = [
      '#type' => 'hidden',
      '#value' => $sdcPluginId,
    ];

    $form['submit'] = [
      '#type' => 'submit',
      // phpcs:ignore
      '#value' => $this->t(self::SUBMIT_BTN_TEXT),
      '#weight' => -30,
    ];

    // Now we build the rest of the form based on the component's props/slots.
    $form['props'][$component_name]['props'] = [
      '#type' => 'markup',
      '#markup' => $this->t('<h3>Props</h3>'),
    ];

    // We loop through all the props and build form elements based on the type
    // of property. For example, an integer can use a textfield, but a string
    // could well need to be a WYSIWYG and properties with enumerated (enum)
    // values (e.g. foo, bar, baz) can be thrown into a select list.
    if (!empty($this->activeComponent->metadata->schema['properties'])) {
      foreach ($this->activeComponent->metadata->schema['properties'] as $property_name => $prop_schema) {
        $form['props'][$component_name][$property_name] = [
          '#type' => 'fieldset',
          '#value' => $property_name,
        ];

        // Instantiate variable to hold the prop type we expect to work with.
        $props_type = '';
        // The field's type (e.g. "string") defined in yml is represented here
        // as two values in an array (e.g. [string, object]) so it seems a
        // field's type is its type plus "object".
        if (isset($prop_schema['type']) && is_array($prop_schema['type'])) {
          // Use array_shift() to get the value we really want. (E.g. "string").
          $props_type = array_shift($prop_schema['type']);
        }
        // Add the property type as a hidden input in the form.
        $form['props'][$component_name][$property_name]['type'] = [
          '#type' => 'hidden',
          '#value' => $props_type,
        ];
        $form['props'][$component_name][$property_name]['type_markup'] = [
          '#type' => 'markup',
          '#markup' => $this->t('<p>Type: :props_type</p>', [':props_type' => $props_type]),
        ];

        // If the component does not define defaults for its properties, use
        // fallbacks so we have something to render.
        if (!isset($prop_schema['default'])) {
          $prop_schema['default'] = $this->componentManager::PROPERTY_TYPE_DEFAULTS[$props_type] ?? $this->t('No default value available');
        }

        // Use our sdc_library.component_manager service to grab values.
        $default_value = $this->componentManager->getPropertyValue($props_type, $prop_schema['default']);

        // The form element to use per type is a mish mash, map it here.
        $form_element_and_format_mapping = [
          'array' => [
            'type' => 'textarea',
            'format' => 'plain_text',
            'default_value' => json_encode($default_value),
          ],
          'boolean' => [
            'type' => 'textfield',
            'format' => 'plain_text',
            'default_value' => $default_value,
          ],
          'integer' => [
            'type' => 'textfield',
            'format' => 'plain_text',
            'default_value' => $default_value,
          ],
          'number' => [
            'type' => 'textfield',
            'format' => 'plain_text',
            'default_value' => $default_value,
          ],
          'object' => [
            'type' => 'textarea',
            'format' => 'plain_text',
            'default_value' => json_encode($default_value),
          ],
          'string' => [
            'type' => 'text_format',
            'format' => 'full_html',
            'default_value' => $default_value,
          ],
        ];
        // Populate a simply-named variable with the form element type we need.
        $form_element_type = $form_element_and_format_mapping[$props_type]['type'];

        // But wait, what if it's a property with a limited set of enum values?
        // Let's build a select list for this use case.
        if (isset($prop_schema['enum'])) {
          $options = [];
          foreach ($prop_schema['enum'] as $value) {
            $form_element_type = 'select';
            $options[$value] = $value;
          }
        }

        // Now we should know enough to build the correct type of form element.
        // These three properties are common to the form element we use.
        $form['props'][$component_name][$property_name]['default'] = [
          '#type' => $form_element_type,
          '#title' => $this->t('Value'),
          '#default_value' => $form_element_and_format_mapping[$props_type]['default_value'],
        ];
        // Here we check for the type and add additional properties.
        if ($form_element_type == 'select') {
          $form['props'][$component_name][$property_name]['default']['#options'] = $options;
        }
        else {
          $form['props'][$component_name][$property_name]['default']['#format'] = $form_element_and_format_mapping[$props_type]['format'];
          $form['props'][$component_name][$property_name]['default']['#maxlength'] = self::TEXTFIELD_MAXLENGTH;
          $form['props'][$component_name][$property_name]['default']['#size'] = self::TEXTFIELD_SIZE;
        }
        // Some props have "examples", not "default".
        // @todo Should we bother supporting "examples"?
        if (isset($prop_schema['examples'])) {
          foreach ($prop_schema['examples'] as $key => $value) {
            $form['props'][$component_name][$property_name]['examples'][$key] = [
              '#type' => 'markup',
              '#markup' => $this->t('<p>Example :num: :value</p>', [':num' => $key + 1, ':value' => $value]),
            ];
          }
        }
        if (isset($prop_schema['title'])) {
          $form['props'][$component_name][$property_name]['title'] = [
            '#type' => 'markup',
            '#markup' => $this->t('<p>Title: :title</p>', [':title' => $prop_schema['title']]),
          ];
        }
        if (isset($prop_schema['description'])) {
          $form['props'][$component_name][$property_name]['description'] = [
            '#type' => 'markup',
            '#markup' => $this->t('<p>Description: :description</p>', [':description' => $prop_schema['description']]),
          ];
        }
      }
    }

    // Based on the schema, it looks like slots only have Title (string),
    // Description (string), Examples (array). Slots have "examples", not
    // "default"!
    if (!empty($this->activeComponent->metadata->slots)) {

      $form['slots'][$component_name]['slots'] = [
        '#type' => 'markup',
        '#markup' => $this->t('<h3>Slots</h3>'),
      ];

      foreach ($this->activeComponent->metadata->slots as $slot => $info) {
        $form['slots'][$component_name][$slot] = [
          '#type' => 'fieldset',
        ];
        $form['slots'][$component_name][$slot]['data'] = [
          '#type' => 'markup',
          '#markup' => $this->t('<h4>:slot</h4>', [':slot' => $slot]),
        ];
        if (
          isset($info['title'])
          && isset($info['description'])
        ) {
          $form['slots'][$component_name][$slot]['more_info'] = [
            '#type' => 'markup',
            '#markup' => '<ul><li>' . $this->t('Title: :title', [':title' => $info['title']]) . '</li><li>' . $this->t('Description: :description', [':description' => $info['description']]) . '</li></ul>',
          ];
        }
        // Slots use "examples" to hold defaults.
        if (isset($info['examples']) && is_array($info['examples'])) {
          foreach ($info['examples'] as $key => $value) {
            $form['slots'][$component_name][$slot][$key] = [
              '#type' => 'text_format',
              '#format' => 'full_html',
              '#title' => $this->t('Editable value'),
              '#default_value' => $value,
              '#maxlength' => self::TEXTFIELD_MAXLENGTH,
              '#size' => self::TEXTFIELD_SIZE,
            ];
          }
        }
        else {
          // This slot doesn't have an example. Provide a default.
          $form['slots'][$component_name][$slot][] = [
            '#type' => 'text_format',
            '#format' => 'full_html',
            '#title' => $this->t('Editable value'),
            '#default_value' => $this->t('This slot does not have examples to use.'),
            '#maxlength' => self::TEXTFIELD_MAXLENGTH,
            '#size' => self::TEXTFIELD_SIZE,
          ];
        }
      }
    }
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
  }

}
